#include <CHM/CHMtcpipPrecond.h>
#pragma hdrstop

#include <CHM/CHMmfcSocketTransport.h>
#include <CHM/CHMvector.h>

class CHMmfcSocketTransportPrivate 
{
public:
   // by default we give parent settings
   CHMmfcSocketTransportPrivate(size_t     iTransportIndex=0, 
                                CHMboolean iIsActive=false);
   ~CHMmfcSocketTransportPrivate();

   size_t TransportIndex;  // = 0 for the main server socket
                           // = 1,2,3.... for all the children

   CHMboolean IsActive;

   CHMvector<CHMmfcSocketTransport*>* pSocketArray;
   CHMstring Host;
   UINT Port;
};

#define CHM_SOCKET_LOG(x)\
if (Log() != NULL)\
{\
   Log()->OutputLine(x);\
}

CHMmfcSocketTransportPrivate::CHMmfcSocketTransportPrivate
(
   size_t     iTransportIndex,
   CHMboolean iIsActive
) 
 : pSocketArray(NULL), 
   TransportIndex(iTransportIndex),  
   IsActive(iIsActive)
{
   // do nothing
}

CHMmfcSocketTransportPrivate::~CHMmfcSocketTransportPrivate()
{
   if (pSocketArray != NULL)
   {
      for (size_t SocketIndex = 0;
           SocketIndex < pSocketArray->size();
           SocketIndex++)
      {
         delete pSocketArray->operator[](SocketIndex);
      }
      delete pSocketArray;
   }
}

CHMmfcSocketTransport::CHMmfcSocketTransport()
{
   pMember = new CHMmfcSocketTransportPrivate;
}

CHMmfcSocketTransport::~CHMmfcSocketTransport()
{
   delete pMember;
}


void CHMmfcSocketTransport::SendResponseMessage
(
   const CHMstring& MessageData,
   size_t TransportIndex
)
{
   if (TransportIndex == 0)
   {
      // okay we sending on this socket - no need
      // to re-route to a child socket
      SendMessage(MessageData);
   }
   else
   {
      // sending to a child socket
      if (pMember->pSocketArray != NULL)
      {
         if (TransportIndex <= pMember->pSocketArray->size())
         {
            pMember->pSocketArray->operator[](TransportIndex-1)->SendResponseMessage
                                                (MessageData, 0);
         }
         else
         {
            OnError(CHMmfcSocketTransport::SOCKET_DOES_NOT_EXIST, CHM_TEXT("Socket does not exist."));
         }
      }
      else
      {
         OnError(CHMmfcSocketTransport::SOCKET_NOT_A_SERVER, CHM_TEXT("This is not a server socket."));
      }
   }
}

void CHMmfcSocketTransport::GetDataWithoutNullChars(CHMstring& Buffer, size_t CountOfChar)
{
   size_t CharIndex;
   for (CharIndex = 0; CharIndex < CountOfChar; CharIndex += 1)
   {
      if (0 == Buffer[CharIndex])
      {
         Buffer[CharIndex] = 32;
         CHMstring LogMessage("Warning! Illegal null character in message at ");
         CHMstring TempBuffer(32,'\0');
         sprintf(TempBuffer.get_buffer(),"%d",CharIndex);
         LogMessage += TempBuffer.c_str();
         CHM_SOCKET_LOG(LogMessage.c_str());
      }
   }
}

BOOL CHMmfcSocketTransport::Connect(LPCTSTR HostAddress, UINT Port)
{
   if ((pMember->Host.compare(HostAddress) == 0) &&
      (pMember->Port == Port) && pMember->IsActive)
   {
      return TRUE;
   }
   else
   {
      pMember->Host = HostAddress;
      pMember->Port = Port;
      if (CSocket::Connect(HostAddress, Port))
      {
         pMember->IsActive = true;
         return TRUE;
      }
      else
      {
         pMember->IsActive = false;
         return FALSE;
      }
   }
}

void CHMmfcSocketTransport::Close()
{
   CSocket::Close();
   pMember->IsActive = false;
}

void CHMmfcSocketTransport::SendMessage
(
   const CHMstring& Data
)
{
   const CHMchar* pBuffer = Data.c_str();
   int BufferLength = (int)Data.size();

   CHMboolean SendComplete = FALSE;

   while (!SendComplete)
   {
      int CharsSent = Send(pBuffer, BufferLength, 0);

      if (CharsSent == SOCKET_ERROR)
      {
         OnError(GetLastError(), CHM_TEXT("Winsock write error."));
         return;
      }
      else
      {
         if (CharsSent < BufferLength)
         {
            pBuffer      += CharsSent;
            BufferLength -= CharsSent;
         }
         else
         {
            SendComplete = TRUE;
         }
      }
   }
}

// Do not edit the following lines, which are needed by ClassWizard.
#if 0
BEGIN_MESSAGE_MAP(CHMsocket, CSocket)
   //{{AFX_MSG_MAP(CHMsocket)
   //}}AFX_MSG_MAP
END_MESSAGE_MAP()
#endif   // 0

const size_t& CHMmfcSocketTransport::TransportIndex() const
{
   return pMember->TransportIndex;
}

void CHMmfcSocketTransport::OnReceive(int nErrorCode) 
{
   if (nErrorCode != 0)
   {
      CHM_SOCKET_LOG("Network subsystem failed.");
   }
   else
   {
      size_t CountOfChar;
      CHMstring IncomingData;
      size_t SizeOfBuffer = 1472;
      CHMstring Buffer;
      Buffer.assign(SizeOfBuffer+1, '\0');
      CountOfChar = Receive((void*)Buffer.get_buffer(), SizeOfBuffer);
      if (CountOfChar != SOCKET_ERROR)
      {
         *(((char*)Buffer.c_str()) + CountOfChar) = 0;
         GetDataWithoutNullChars(Buffer,CountOfChar);   
         IncomingData = IncomingData + Buffer;
      }
      else
      {
         OnError(GetLastError(), CHM_TEXT("Winsock error receiving data."));
      }
      CHM_ASSERT(Wrapper() != NULL);
      Wrapper()->ReceiveData(IncomingData, pMember->TransportIndex);
   }
   CSocket::OnReceive(nErrorCode);
}

void CHMmfcSocketTransport::OnAccept(int nErrorCode) 
{
   if (nErrorCode != 0)
   {
      OnError(GetLastError(), CHM_TEXT("Winsock error accepting connection."));
   }
   else
   {
      CHM_SOCKET_LOG("Received connection...");
      if (pMember->pSocketArray == NULL)
      {
         pMember->pSocketArray = new CHMvector<CHMmfcSocketTransport*>;
      }
      
      size_t SocketIndex = 0;
      while (SocketIndex < pMember->pSocketArray->size()
             && true == pMember->pSocketArray->operator[](SocketIndex)->pMember->IsActive)
      {
         SocketIndex++;
      }
      CHM_ASSERT(SocketIndex <= pMember->pSocketArray->size());
      if (SocketIndex == pMember->pSocketArray->size())
      {
         CHMstring LogMessage("Creating new socket ");
         CHMstring TempBuffer(32,'\0');
         sprintf(TempBuffer.get_buffer(),"%d",SocketIndex + 1);
         LogMessage += TempBuffer.c_str();
         CHM_SOCKET_LOG(LogMessage.c_str());

         pMember->pSocketArray->push_back(
             new CHMmfcSocketTransport);
         pMember->pSocketArray
              ->operator[](pMember->pSocketArray->size()-1)
              ->pMember->IsActive = true;
         pMember->pSocketArray
              ->operator[](pMember->pSocketArray->size()-1)
              ->pMember->TransportIndex = pMember->pSocketArray->size();
         pMember->pSocketArray
              ->operator[](pMember->pSocketArray->size()-1)
              ->SetEngine(Engine());
      }
      else
      {
         CHMstring LogMessage("Reusing socket ");
         CHMstring TempBuffer(32,'\0');
         sprintf(TempBuffer.get_buffer(),"%d",SocketIndex + 1);
         LogMessage += TempBuffer.c_str();
         CHM_SOCKET_LOG(LogMessage.c_str());
         
         pMember->pSocketArray->operator [](SocketIndex)->Detach();
         pMember->pSocketArray->operator [](SocketIndex)->pMember->IsActive = true;
      }
      CHM_ASSERT(Wrapper() != NULL);
      // clear buffer of any existing data from previous connection
      Wrapper()->ClearData(SocketIndex);
      
      Accept(*pMember->pSocketArray->operator[](SocketIndex));
   }
   
   CSocket::OnAccept(nErrorCode);
}

void CHMmfcSocketTransport::OnClose(int nErrorCode) 
{   
   pMember->IsActive = false;

   CHMstring LogMessage("Closed socket ");
   CHMstring TempBuffer(32,'\0');
   sprintf(TempBuffer.get_buffer(),"%d",pMember->TransportIndex);
   LogMessage += TempBuffer.c_str();
   CHM_SOCKET_LOG(LogMessage.c_str());
   
   CSocket::OnClose(nErrorCode);
}
